/* $Header: /ext/yav/catty/xpx/RCS/convert.c,v 1.9 1994/01/17 22:19:08 yav Exp $
 */

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <stdio.h>

#include "xpx.h"
#include "extern.h"
#define col3to8(x) (((x)<<5)|((x)<<2)|((x)>>1))

/* return : read color set */
read_ccb(fp, p)
     FILE *fp; COL *p;
{
  unsigned char buf[2*MAXCOLSET];
  int i, r;

  r = 0;
  while (fread(buf, 2*MAXCOLSET, 1, fp) != 0) {
    for (i = 0; i < MAXCOLSET; i++) {
      p->blue = buf[i*2] & 7;
      p->red = (buf[i*2]>>3) & 7;
      p->green = ((buf[i*2]>>6) & 3)|((buf[i*2+1]&1) << 2);
      p->blue = col3to8(p->blue);
      p->red = col3to8(p->red);
      p->green = col3to8(p->green);
      p++;
    }
    r++;
  }
  return r;
}

/* return pnm type number */
read_pnm_header(fp, data, colorhints, nhints)
     FILE *fp;
     int *data;
     COL *colorhints;
     int *nhints;
{
  int i, result, r, g, b;
  int datan, maxcol;
  char buf[256], *p;

  if (fgets(buf, sizeof(buf), fp) == NULL)
    return 0;
  if (buf[0] != 'P') {
    fprintf(stderr, "not ppm file!\n");
    return 0;
  }
  switch (buf[1]) {
  case '2':
    result = 2;
    break;
  case '3':
    result = 3;
    break;
  case '5':
    result = 5;
    break;
  case '6':
    result = 6;
    break;
  default:
    fprintf(stderr, "not supported ppm,\n");
    return 0;
  }
  maxcol = -1;
  datan = 0;
  while (datan < 3 && fgets(buf, sizeof(buf), fp) != NULL) {
    if (buf[0] == '#') {
      if (colorhints != NULL && nhints != NULL) {
	if (sscanf(buf, "# color:%d %d %d %d", &i, &r, &g, &b) == 4) {
	  if (i < *nhints) {
	    (colorhints+i)->red = r;
	    (colorhints+i)->green = g;
	    (colorhints+i)->blue = b;
	    if (maxcol < i)
	      maxcol = i;
	  }
	}
      }
    } else {
      for (p = buf; (i = strtol(p, &p, 10)) != 0 && datan < 3; datan++)
	data[datan] = i;
    }
  }
  if (colorhints != NULL && nhints != NULL)
    *nhints = maxcol+1;
  return result;
}

read_pnm_pixels(fp, t, cp, n)
     FILE *fp;
     int t;
     COL *cp;
     int n;
{
  int i, x, rgbsw;
  char *p, *op;
  char buf[256];

  rgbsw = 0;
  switch(t) {
  case 3:
    p = buf;
    *p = '\0';
    for (x = 0; x < n; x++) {
      if (*p == '\0' || *p == '\n') {
	if (fgets(buf, sizeof(buf), fp) == NULL)
	  return x;
	p = buf;
	continue;
      }
      op = p;
      i = strtol(p, &p, 10);
      if (p == op)
	continue;
      switch(rgbsw) {
      case 0:
	cp->red = i;
	rgbsw = 1;
	break;
      case 1:
	cp->green = i;
	rgbsw = 2;
	break;
      case 2:
	cp->blue = i;
	cp++;
	rgbsw = 0;
	break;
      }
    }
    break;
  case 6:
    return fread(cp, 3, n, fp);
  default:
    fprintf(stderr, "pnm type %d!\n", t);
    return 0;
  }
  /* NOT REACHED */
}

color_pixel(tbl, n, col)
     COL *tbl;
     int n;
     COL *col;
{
 int i;
 
 for (i = 0; i < n; i++) {
   if (col->red == tbl->red &&
       col->green == tbl->green &&
       col->blue == tbl->blue)
     return i;
   tbl++;
 }
 return -1;
}

read_pnm(fp)
     FILE *fp;
{
  int i, n, f, x, y, pnmtype;
  int data[3], w, h;
  int colcnt;
  COL px[640];
  
  colcnt = (MAXPALSET-current_pal)*16;
  pnmtype = read_pnm_header(fp, data, color_buf+16*current_pal, &colcnt);
  if (pnmtype == 0)
    return 1;
  free_color();
  /* magtoppm commented color information set */
  get_color(color_buf+16*current_pal, colcnt);

  w = data[0];
  h = data[1];
  if (w > MAXIMGW) {
    w = MAXIMGW;
    fprintf(stderr, "width %d trancate to %d.\n", data[0], w);
  }
  if (h > MAXIMGH) {
    h = MAXIMGH;
    fprintf(stderr, "height %d trancate to %d.\n", data[1], h);
  }
  imgfilew = w;
  imgfileh = h;
  for (y = 0; y < h; y++) {
    f = 1;
    for (n = data[0]; n; n -= i) {
      i = n > w ? w : n;
      read_pnm_pixels(fp, pnmtype, px, i);
      if (f) {
	for (x = 0; x < i; x++)
	  *(imgdata+y*imgwidth+x) =
	    color_pixel(color_buf+16*current_pal, colcnt, &px[x]);
	f = 0;
      }
    }
  }
  image_update_all();
  return 0;
}

static char kiss_magic_number[] = "KiSS";

read_kcf(fp)
     FILE *fp;
{
  int i, s, bpc, cc, col, set, preload;
  COL *cp;
  unsigned char *p, buf[4096];

  /* old style kcf spec set */
  bpc = 12;
  col = 16;
  set = 10;
  preload = strlen(kiss_magic_number);
  if (!fread(buf, preload, 1, fp))
    return 1;
  if (strncmp(buf, kiss_magic_number, preload)==0) {
    /* read new style kcf header section */
    if (!fread(buf+preload, 32-preload, 1, fp))
      return 1;
    /* check color file mark */
    if (buf[4] != 0x10)
      return 1;
    bpc = buf[5];
    col = buf[8]|(buf[9]<<8);
    set = buf[10]|(buf[11]<<8);
    preload = 0;
  }
  cp = color_buf;
  cc = ((bpc+7)>>3)*col;
  for (s = 0; s < set; s++) {
    if (!fread(buf+preload, cc-preload, 1, fp))
      return 1;
    p = buf;
    for (i = 0; i < col; i++) {
      if (bpc == 12) {
	cp->red = (*p&0xf0)|(*p>>4);
	cp->blue = (*p&0x0f)|((*p&0x0f)<<4);
	p++;
	cp->green = (*p&0x0f)|((*p&0x0f)<<4);
	p++;
      } else {
	cp->red = *p++;
	cp->green = *p++;
	cp->blue = *p++;
      }
      cp++;
    }
    preload = 0;
  }
  free_color();
  get_color(color_buf+16*current_pal, col);
  image_update_all();
  return 0;
}

write_kcf(fp, newver)
     FILE *fp;
{
  int i, s, bpc, col, set, cc;
  COL *cp;
  unsigned char *p, buf[4096];

  /* old style kcf spec set */
  bpc = 12;
  col = 16;
  set = 10;
  if (newver) {
    bpc = 24;
    memset(buf, 0, 32);
    memcpy(buf, kiss_magic_number, strlen(kiss_magic_number));
    /* set color file mark */
    buf[4] = 0x10;
    buf[5] = bpc;
    buf[8] = col;
    buf[9] = col>>8;
    buf[10] = set;
    buf[11] = set>>8;
    if (!fwrite(buf, 32, 1, fp))
      return 1;
  }
  cp = color_buf;
  cc = ((bpc+7)>>3)*col;
  for (s = 0; s < set; s++) {
    p = buf;
    for (i = 0; i < col; i++) {
      if (bpc == 12) {
	*p++ = (cp->red & 0xf0)|(cp->blue >> 4);
	*p++ = cp->green >> 4;
      } else {
	*p++ = cp->red;
	*p++ = cp->green;
	*p++ = cp->blue;
      }
      cp++;
    }
    if (!fwrite(buf, cc, 1, fp))
      return 1;
  }
  return 0;
}

read_cel(fp)
     FILE *fp;
{
  int c, x, y, w, h, preload, bpp;
  unsigned char buf[4096];

  preload = strlen(kiss_magic_number);
  if (!fread(buf, preload, 1, fp))
    return 1;
  if (strncmp(buf, kiss_magic_number, preload) == 0) {
    /* read new style cel file header section */
    if (!fread(buf+preload, 32-preload, 1, fp))
      return 1;
    /* check cel file mark */
    if (buf[4] != 0x20)
      return 1;
    bpp = buf[5];
    w = buf[8]|(buf[9]<<8);
    h = buf[10]|(buf[11]<<8);
  } else {
    bpp = 4;
    w = buf[0]|(buf[1]<<8);
    h = buf[2]|(buf[3]<<8);
  }
  for (y = 0; y < h; y++) {
    if (!fread(buf, bpp==4?((w+1)>>1):w, 1, fp))
      return 1;
    for (x = 0; x < w; x++) {
      if (bpp == 4) {
	c = buf[x>>1];
	if ((x & 1) == 0)
	  c >>= 4;
	else
	  c &= 0x0f;
      } else {
	c = buf[x];
      }
      *(imgdata+y*imgwidth+x) = c;
    }
  }
  imgfilew = w;
  imgfileh = h;
  image_update_all();
  return 0;
}

write_cel(fp, newver)
     FILE *fp;
{
  int c, x, y, w, h, bpp;
  unsigned char buf[4096];

  w = imgfilew;
  h = imgfileh;
  bpp = 4;
  if (newver) {
    memset(buf, 0, 32);
    memcpy(buf, kiss_magic_number, strlen(kiss_magic_number));
    /* set cel file mark */
    buf[4] = 0x20;
    buf[5] = bpp;
    buf[8] = w;
    buf[9] = w >> 8;
    buf[10] = h;
    buf[11] = h >> 8;
    if (!fwrite(buf, 32, 1, fp))
      return 1;
  } else {
    buf[0] = w;
    buf[1] = w >> 8;
    buf[2] = h;
    buf[3] = h >> 8;
    if (!fwrite(buf, 4, 1, fp))
      return 1;
  }
  for (y = 0; y < h; y++) {
    for (x = 0; x < w; x++) {
      c = *(imgdata+y*imgwidth+x);
      if (bpp == 4) {
	if ((x & 1) == 0)
	  buf[x>>1] = c<<4;
	else
	  buf[x>>1] |= c;
      } else {
	buf[x] = c;
      }
    }
    if (!fwrite(buf, bpp==4?((w+1)>>1):w, 1, fp))
      return 1;
  }
  return 0;
}

gen_fontpix(p, cbb)
     unsigned char *p; unsigned char *cbb;
{
  int x, y;
  static unsigned char bittbl[8] = {0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01};

  for (y = 0; y < 8; y++) {
    for (x = 0; x < 8; x++) {
      if (*cbb & bittbl[x])
	*p = 1;
      else
	*p = 0;
      if (*(cbb+1) & bittbl[x])
	*p |= 2;
      if (*(cbb+16) & bittbl[x])
	*p |= 4;
      if (*(cbb+17) & bittbl[x])
	*p |= 8;
      p++;
    }
    cbb += 2;
  }
}

gen_image_cbb(cbb)
     unsigned char *cbb;
{
  int dx, dy, xx, yy, ix, iy, x, y;
  unsigned char fontpix[8*8], *p;
  int c;

  for (y = 0; y < 8*16; y+=8) {
    for (x = 0; x < 8*16; x+=8) {
      gen_fontpix(fontpix, cbb);
      cbb += 32;
      p = fontpix;
      for (iy = 0; iy < 8; iy++) {
	for (ix = 0; ix < 8; ix++) {
	  *(imgdata+(y+iy)*imgwidth+x+ix) = *p++;
	}
      }
    }
  }
}

/* set XImage from imgdata */
setpixel_sub(x0, y0, pixel)
{
  int x, y, x1, y1;

  x0 *= zoomfactor;
  y0 *= zoomfactor;
  x1 = x0+zoomfactor;
  y1 = y0+zoomfactor;
  for (y = y0; y < y1; y++)
    for (x = x0; x < x1; x++)
      XPutPixel(img, x, y, pixel);
}
  
/* set XImage from imgdata */
setpixelcolor(x, y, col)
{
  *(imgdata+y*imgwidth+x) = col;
  setpixel_sub(x, y, pixtbl[col]);
}
  
/* set XImage from imgdata */
gen_image_sub(x0, y0, w, h)
{
  int x, y;
  unsigned char *ip;

  for (y = y0; y < y0+h; y++) {
    ip = imgdata+y*imgwidth+x0;
    for (x = x0; x < x0+w; x++) {
      setpixel_sub(x, y, pixtbl[*ip++]);
    }
  }
}
  
/* End of file */
