/* $Header: /ext/yav/catty/xpx/RCS/file.c,v 1.14 1994/01/17 22:19:08 yav Exp $
 */

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <stdio.h>

#include "xpx.h"
#include "extern.h"

#define BDW 1
#define MAXFILE 1024

#define PATHX 8
#define PATHY 8
#define PATHW 300
#define PATHH 18

#define NAMEX 160
#define NAMEY 100
#define NAMEW 140
#define NAMEH 20

#define DIRSLOTMAX 12
#define DIRSLOTX 8
#define DIRSLOTY 64
#define DIRSLOTW 128
#define DIRSLOTH 18

#define SCROLLX DIRSLOTX
#define SCROLLY (DIRSLOTY-(SCROLLH+BDW))
#define SCROLLW DIRSLOTW
#define SCROLLH 20

#define FILEMENUN 6
#define FILEMENUX 196
#define FILEMENUY 160
#define FILEMENUW 64
#define FILEMENUH 20

#define IMGNMX IMGX
#define IMGNMY (IMGY-40)
#define COLNMX PALSETX
#define COLNMY (PALSETY-24)

static Window filewin;
static Window dirslotwin[DIRSLOTMAX];
static Window scrollwin[2];
static Window filemenuwin[FILEMENUN];
static Window pathwin;
static Window namewin;
static int filewin_opened = 0;
static int select_slot;
static char filenamebuf[MAXFILE][64];
static char current_path[1024];
static char noname[] = "*Noname*";
static char current_name[256];
static int nametop;
static int scroll_dir = 0;
static Window imgfilewin, colfilewin;

extern void menu_stat_change();

get_current_path()
{
  getwd(current_path);
}

get_filenames()
{
  int i, n;
  FILE *fp;
  int fd[2];
  char buf[1024];

  select_slot = nametop = 0;
  pipe(fd);
  i = fork();
  if (i == -1) {
    fprintf(stderr, "fork error!\n");
    return;
  }
  if (i == 0) {
    close(1);
    dup(fd[1]);
    close(fd[1]);
    close(fd[0]);
    execlp("ls", "ls", "-aF", current_path, NULL);
    fprintf(stderr, "exec ls error!\n");
    exit(1);
  }
  fp = fdopen(fd[0], "r");
  close(fd[1]);
  n = 0;
  while (fgets(buf, sizeof(buf), fp) != NULL) {
    i = strlen(buf);
    if (i && buf[i-1] == '\n')
      buf[i-1] = '\0';
    if (n >= nametop) {
      strncpy(filenamebuf[n], buf, sizeof(*filenamebuf)-1);
      filenamebuf[n][sizeof(*filenamebuf)-1] = '\0';
    }
    n++;
    if (n >= MAXFILE)
      break;
  }
  while (n-nametop < MAXFILE) {
    filenamebuf[n][0] = '\0';
    n++;
  }
  fclose(fp);
  wait(&i);
  XClearArea(dsp, pathwin, 0, 0, 0, 0, True);
  for (i = 0; i < DIRSLOTMAX; i++)
    XClearArea(dsp, dirslotwin[i], 0, 0, 0, 0, True);
}

create_file_window(x, y)
{
  int i;
  XSizeHints hint;
  char buf[256];

  if (filewin_opened) {
    XMapWindow(dsp, filewin);
    return 1;
  }
  hint.x = x; hint.y = y;
  hint.width = 320; hint.height = 320;
  hint.flags = PPosition|PSize;
  filewin = XCreateSimpleWindow(dsp, DefaultRootWindow(dsp), hint.x, hint.y,
				hint.width, hint.height, BDW, fg, bg);
  sprintf(buf, "%s file", myname);
  XSetStandardProperties(dsp, filewin, buf, buf, None, NULL, NULL, &hint);
  XMapRaised(dsp, filewin);

  pathwin = XCreateSimpleWindow(dsp, filewin, PATHX, PATHY, PATHW, PATHH,
				BDW, fg, bg);
  XSelectInput(dsp, pathwin,
	       ButtonPressMask|ExposureMask|EnterWindowMask|LeaveWindowMask);
  XMapRaised(dsp, pathwin);

  namewin = XCreateSimpleWindow(dsp, filewin, NAMEX, NAMEY, NAMEW, NAMEH,
				BDW, fg, bg);
  XSelectInput(dsp, namewin,
	       ButtonPressMask|ExposureMask|EnterWindowMask|LeaveWindowMask);
  XMapRaised(dsp, namewin);

  x = FILEMENUX;
  y = FILEMENUY;
  for (i = 0; i < FILEMENUN; i++) {
    filemenuwin[i] = XCreateSimpleWindow(dsp, filewin, x, y,
					 FILEMENUW, FILEMENUH,
					 BDW, fg, bg);
    XSelectInput(dsp, filemenuwin[i],
		 ButtonPressMask|ExposureMask|EnterWindowMask|LeaveWindowMask);
    XMapRaised(dsp, filemenuwin[i]);
    y += FILEMENUH+BDW;
  }
  x = DIRSLOTX;
  y = DIRSLOTY;
  for (i = 0; i < DIRSLOTMAX; i++) {
    dirslotwin[i] = XCreateSimpleWindow(dsp, filewin, x, y, DIRSLOTW, DIRSLOTH,
					0, fg, bg);
    XSelectInput(dsp, dirslotwin[i],
		 ButtonPressMask|ExposureMask|EnterWindowMask|LeaveWindowMask);
    XMapRaised(dsp, dirslotwin[i]);
    y += DIRSLOTH+0;
  }
  x = SCROLLX;
  y = SCROLLY;
  for (i = 0; i < 2; i++) {
    scrollwin[i] = XCreateSimpleWindow(dsp, filewin, x, y, SCROLLW, SCROLLH,
				       BDW, fg, bg);
    XSelectInput(dsp, scrollwin[i],ExposureMask|
		 ButtonPressMask|ButtonReleaseMask|OwnerGrabButtonMask|
		 EnterWindowMask|LeaveWindowMask);
    XMapRaised(dsp, scrollwin[i]);
    y += SCROLLH+BDW+(DIRSLOTH+0)*DIRSLOTMAX;
  }
  get_current_path();
  get_filenames();
  filewin_opened = 1;
  return 0;
}

unmap_file_window()
{
  XUnmapWindow(dsp, filewin);
}

close_file_window()
{
  int i;

  if (!filewin_opened)
    return;
  filewin_opened = 0;
  for (i = 0; i < DIRSLOTMAX; i++)
    XDestroyWindow(dsp, dirslotwin[i]);
  for (i = 0; i < FILEMENUN; i++)
    XDestroyWindow(dsp, filemenuwin[i]);
  XDestroyWindow(dsp, scrollwin[0]);
  XDestroyWindow(dsp, scrollwin[1]);
  XDestroyWindow(dsp, namewin);
  XDestroyWindow(dsp, pathwin);
  XDestroyWindow(dsp, filewin);
}

change_parent_dir()
{
  char *p;
  extern char *strrchr();
  
  p = strrchr(current_path, '/');
  if (p != NULL && p != current_path)
    *p = '\0';
}

void set_imgfilename(name)
     char *name;
{
  strcpy(imgfilename, name);
  XClearArea(dsp, imgfilewin, 0, 0, 0, 0, True);
  update_infowin();
}

void set_colfilename(name)
     char *name;
{
  strcpy(colfilename, name);
  XClearArea(dsp, colfilewin, 0, 0, 0, 0, True);
  update_infowin();
}

create_filename_window()
{
  imgfilewin = XCreateSimpleWindow(dsp, win, IMGNMX, IMGNMY, 100, 18, 1, fg, bg);
  XSelectInput(dsp, imgfilewin, ExposureMask);
  XMapRaised(dsp, imgfilewin);
  colfilewin = XCreateSimpleWindow(dsp, win, COLNMX, COLNMY, 100, 18, 1, fg, bg);
  XSelectInput(dsp, colfilewin, ExposureMask);
  XMapRaised(dsp, colfilewin);
}

close_filename_window()
{
  XDestroyWindow(dsp, colfilewin);
  XDestroyWindow(dsp, imgfilewin);
}

filename_expose(ev)
     XEvent *ev;
{
  extern char *filname();
  
  if (ev->xexpose.window == imgfilewin) {
    menuwin_center(imgfilewin, filename(imgfilename), 0);
    return;
  }
  if (ev->xexpose.window == colfilewin) {
    menuwin_center(colfilewin, filename(colfilename), 0);
    return;
  }
}

read_cbb_file(name)
     char *name;
{
  FILE *fp;
  int r;
  unsigned char cbb_data[32*256];
  
  fp = fopen(name, "r");
  if (fp == NULL) {
    fprintf(stderr, "cbb file ``%s'' open error!\n", name);
    return 1;
  }
  set_imgfilename(name);
  init_image();
  r = fread(cbb_data, sizeof(cbb_data), 1, fp);
  fclose(fp);
  if (!r) {
    set_imgfilename("");
    fprintf(stderr, "cbb file ``%s'' open error!\n", name);
    return 1;
  }
  imgfilew = 128;
  imgfileh = 128;
  gen_image_cbb(cbb_data);
  gen_image_sub(0, 0, imgfilew, imgfileh);
  img_expose(0, 0, imgwinwidth, imgwinheight);
  return 0;
}

read_ccb_file(name)
     char *name;
{
  FILE *fp;
  int r;
  
  fp = fopen(name, "r");
  if (fp == NULL) {
    fprintf(stderr, "ccb file ``%s'' open error!\n", name);
    return 1;
  }
  set_colfilename(name);
  r = read_ccb(fp, color_buf);
  fclose(fp);
  if (r == 0) {
    set_colfilename("");
    fprintf(stderr, "ccb file ``%s'' read error!\n", name);
    return 1;
  }
  free_color();
  get_color(color_buf+current_pal*16, 16);
  gen_image_sub(0, 0, imgfilew, imgfileh);
  img_expose(0, 0, imgwinwidth, imgwinheight);
  return 0;
}

read_mag_file(name)
     char *name;
{
  FILE *fp;
  int r, fd[2];
  
  pipe(fd);
  if (fork() == 0) {
    close(1);
    dup(fd[1]);
    close(fd[1]);
    close(fd[0]);
    execlp("magtoppm", "magtoppm", name, NULL);
    exit(0);
  }
  fp = fdopen(fd[0], "r");
  close(fd[1]);
  set_colfilename(name);
  set_imgfilename(name);
  r = read_pnm(fp);
  fclose(fp);
  if (r) {
    set_colfilename("");
    set_imgfilename("");
    fprintf(stderr, "mag file ``%s'' read error!\n", name);
  }
  return r;
}

read_pnm_file(name)
     char *name;
{
  FILE *fp;
  int r;

  fp = fopen(name, "r");
  if (fp == NULL) {
    fprintf(stderr, "pnm file ``%s'' open error!\n", name);
    return 1;
  }
  set_colfilename(name);
  set_imgfilename(name);
  r = read_pnm(fp);
  fclose(fp);
  if (r) {
    set_colfilename("");
    set_imgfilename("");
    fprintf(stderr, "pnm file ``%s'' read error!\n", name);
  }
  return r;
}

read_kcf_file(name)
     char *name;
{
  FILE *fp;
  int r;

  fp = fopen(name, "r");
  if (fp == NULL) {
    fprintf(stderr, "kcf file ``%s'' open error!\n", name);
    return 1;
  }
  set_colfilename(name);
  r = read_kcf(fp);
  fclose(fp);
  if (r) {
    set_colfilename("");
    fprintf(stderr, "kcf file ``%s'' read error!\n", name);
    return 1;
  }
  return 0;
}

read_cel_file(name)
     char *name;
{
  FILE *fp;
  int r;
  
  fp = fopen(name, "r");
  if (fp == NULL) {
    fprintf(stderr, "cel file ``%s'' open error!\n", name);
    return 1;
  }
  set_imgfilename(name);
  r = read_cel(fp);
  fclose(fp);
  if (r) {
    set_imgfilename("");
    fprintf(stderr, "cel file ``%s'' read error!\n", name);
    return 1;
  }
  return 0;
}

struct {
  char *suffix;
  int (*func)();
} file_type_tbl[] = {
  {".ccb", read_ccb_file},
  {".cbb", read_cbb_file},
  {".ppm", read_pnm_file},
  {".mag", read_mag_file},
  {".kcf", read_kcf_file},
  {".cel", read_cel_file},
  {NULL, NULL}
};

write_kcf_file(name)
     char *name;
{
  FILE *fp;
  int r;

  fp = fopen(name, "w");
  if (fp == NULL) {
    fprintf(stderr, "kcf file ``%s'' write open error!\n", name);
    return 1;
  }
  r = write_kcf(fp, 1);
  fclose(fp);
  if (r) {
    fprintf(stderr, "kcf file ``%s'' write error!\n", name);
    return 1;
  }
  return 0;
}


write_cel_file(name)
     char *name;
{
  FILE *fp;
  int r;

  fp = fopen(name, "w");
  if (fp == NULL) {
    fprintf(stderr, "cel file ``%s'' write open error!\n", name);
    return 1;
  }
  r = write_cel(fp, 1);
  fclose(fp);
  if (r) {
    fprintf(stderr, "cel file ``%s'' write error!\n", name);
    return 1;
  }
  return 0;
}


struct {
  char *suffix;
  int (*func)();
} save_type_tbl[] = {
  {".kcf", write_kcf_file},
  {".cel", write_cel_file},
  {NULL, NULL}
};

set_current_name(name)
     char *name;
{
  strcpy(current_name, name);
  XClearArea(dsp, namewin, 0, 0, 0, 0, True);
}

reset_current_name()
{
  set_current_name("");
}

read_data_file(name)
     char *name;
{
  int i, n;
  
  for (n = 0; file_type_tbl[n].suffix != NULL; n++) {
    i = strlen(name) - strlen(file_type_tbl[n].suffix);
    if (i >= 0 && strcmp(name+i, file_type_tbl[n].suffix) == 0) {
      file_type_tbl[n].func(name);
      return 0;
    }
  }
  return 1;
}

file_open()
{
  int i;
  char buf[1024];

  i = strlen(current_name);
  if (i && current_name[i-1] == '/') {
    current_name[i-1] = '\0';
    if (strcmp(current_name, "..") == 0)
      change_parent_dir();
    else if (strcmp(current_name, ".")) {
      strcat(current_path, "/");
      strcat(current_path, current_name);
    }
    get_filenames();
    reset_current_name();
  }
  if (current_name[0] == '\0')
    return;
  sprintf(buf, "%s/%s", current_path, current_name);
  return read_data_file(buf);
}

file_save()
{
  int i, n;
  char buf[1024];

  i = strlen(current_name);
  if (i && current_name[i-1] == '/')
    return;
  if (current_name[0] == '\0')
    return;
  sprintf(buf, "%s/%s", current_path, current_name);
  for (n = 0; save_type_tbl[n].suffix != NULL; n++) {
    i = strlen(buf) - strlen(save_type_tbl[n].suffix);
    if (i >= 0 && strcmp(buf+i, save_type_tbl[n].suffix) == 0) {
      save_type_tbl[n].func(buf);
      return;
    }
  }
}

file_remove()
{
  int i, n;
  char buf[1024];

  i = strlen(current_name);
  if (i && current_name[i-1] == '/')
    return;
  if (current_name[0]) {
    sprintf(buf, "%s/%s", current_path, current_name);
    unlink(buf);
    get_filenames();
    reset_current_name();
  }
}

redraw_dirslot(n)
{
  menuwin_left(dirslotwin[n], filenamebuf[nametop+n], nametop+n==select_slot);
}

redraw_filemenu(n)
{
  static char *menustr[FILEMENUN] = {"Open", "Save", "----", "Remove",
				       "----", "Cancel"};
  
  menuwin_center(filemenuwin[n], menustr[n], 0);
}

file_expose(ev)
     XEvent *ev;
{
  int i;

  if (!filewin_opened)
    return;
  if (ev->xexpose.count != 0)
    return;
  if (ev->xexpose.window == namewin) {
    menuwin_center(namewin, current_name[0]?current_name:noname, 0);
    return;
  }
  if (ev->xexpose.window == pathwin) {
    menuwin_center(pathwin, current_path, 0);
    return;
  }
  if (ev->xexpose.window == scrollwin[0]) {
    menuwin_center(scrollwin[0], " ^ ", scroll_dir==-1);
    return;
  }
  if (ev->xexpose.window == scrollwin[1]) {
    menuwin_center(scrollwin[1], " v ", scroll_dir==1);
    return;
  }
  for (i = 0; i < FILEMENUN; i++) {
    if (ev->xexpose.window == filemenuwin[i]) {
      redraw_filemenu(i);
      return;
    }
  }
  for (i = 0; i < DIRSLOTMAX; i++) {
    if (ev->xexpose.window == dirslotwin[i]) {
      redraw_dirslot(i);
      return;
    }
  }
}

scroll_redraw()
{
  int i;
  for (i = 0; i < DIRSLOTMAX; i++)
    XClearArea(dsp, dirslotwin[i], 0, 0, 0, 0, True);
}

scroll_down()
{
  if (nametop) {
    --nametop;
    scroll_redraw();
  }
}

scroll_up()
{
  if (filenamebuf[nametop+DIRSLOTMAX-1][0] != '\0') {
    ++nametop;
    scroll_redraw();
  }
}

file_button(ev)
     XEvent *ev;
{
  int i, oldsel;

  if (!filewin_opened)
    return;
  if (ev->xbutton.window == scrollwin[0]) {
    scroll_dir = -1;
    XClearArea(dsp, scrollwin[0], 0, 0, 0, 0, True);
    scroll_down();
    return;
  }
  if (ev->xbutton.window == scrollwin[1]) {
    scroll_dir = 1;
    XClearArea(dsp, scrollwin[1], 0, 0, 0, 0, True);
    scroll_up();
    return;
  }
  for (i = 0; i < DIRSLOTMAX; i++) {
    if (ev->xbutton.window == dirslotwin[i]) {
      switch(ev->xbutton.button) {
      case 1:
	if (filenamebuf[nametop+i][0] == '\0')
	  return;
	oldsel = select_slot;
	select_slot = nametop+i;
	set_current_name(filenamebuf[select_slot]);
	if (oldsel >= nametop && oldsel < nametop+DIRSLOTMAX)
	  XClearArea(dsp, dirslotwin[oldsel-nametop], 0, 0, 0, 0, True);
	XClearArea(dsp, dirslotwin[select_slot-nametop], 0, 0, 0, 0, True);
	if (double_clicked)
	  file_open();
	return;
      }
    }
  }
  for (i = 0; i < FILEMENUN; i++) {
    if (ev->xbutton.window == filemenuwin[i]) {
      switch(i) {
      case 0:
	file_open();
	break;
      case 1:
	file_save();
	break;
      case 2:
	break;
      case 3:
	file_remove();
	break;
      case 4:
	break;
      case 5:
	menu_stat_change(1);
	break;
      }
      return;
    }
  }
}

file_release(ev)
     XEvent *ev;
{
  if ((ev->xbutton.window == scrollwin[0])||
      (ev->xbutton.window == scrollwin[1])) {
    scroll_dir = 0;
    XClearArea(dsp, scrollwin[0], 0, 0, 0, 0, True);
    XClearArea(dsp, scrollwin[1], 0, 0, 0, 0, True);
  }
}

file_enter(ev)
     XEvent *ev;
{
  int i;

  if (!filewin_opened)
    return;
  if (ev->xcrossing.window == pathwin) {
    menuwin_highlight(pathwin, 0);
    return;
  }
  if (ev->xcrossing.window == namewin) {
    menuwin_highlight(namewin, 0);
    return;
  }
  if (ev->xcrossing.window == scrollwin[0]) {
    menuwin_highlight(scrollwin[0], 0);
    return;
  }
  if (ev->xcrossing.window == scrollwin[1]) {
    menuwin_highlight(scrollwin[1], 0);
    return;
  }
  for (i = 0; i < FILEMENUN; i++) {
    if (ev->xcrossing.window == filemenuwin[i]) {
      menuwin_highlight(filemenuwin[i], 0);
      return;
    }
  }
  for (i = 0; i < DIRSLOTMAX; i++) {
    if (ev->xcrossing.window == dirslotwin[i]) {
      menuwin_highlight(dirslotwin[i], nametop+i == select_slot);
      return;
    }
  }
}

file_background()
{
  switch (scroll_dir) {
  case 1:
    scroll_up();
    break;
  case -1:
    scroll_down();
    break;
  }
}

/* End of file */
