/* $Header: /ext/yav/catty/xpx/RCS/main.c,v 1.23 1994/01/17 22:19:08 yav Exp $
 */

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <stdio.h>

#include "xpx.h"

#define IMGSCRLW 16

#define COLSETX (PALSETX+32)
#define COLSETY PALSETY

#define IMGWINWIDTHMAX 640
#define IMGWINHEIGHTMAX 800

extern Window cursorwin[];

char rcsrev[] = "$Revision: 1.23 $";
char *myname = "xpx";
Display *dsp;
int scr;
Window win;
unsigned long fg, bg;
GC gc0, gc1, gridgc;
XImage *img;
Window imgwin;
XFontStruct *fs0, *fs1;
Window colwin[MAXCOLSET];
GC colgc[MAXCOLSET];
int current_col = -1;
int current_pal = 0;
int sw_quit = 0;
int double_clicked = 0;
int zoomfactor = 2;
char imgfilename[1024] = "*Noname*";
char colfilename[1024] = "*Noname*";
unsigned long pixtbl[MAXCOLSET];
unsigned char *imgdata;
int imgwidth, imgheight;
int imgofsx = 0;
int imgofsy = 0;
int imgwinwidth, imgwinheight;
int alloced_color = 0;
int grid_mode = 0;
int grid_size = 8;
int imgfilew = 0;
int imgfileh = 0;
int curx = 0;
int cury = 0;
int curz = 0;

static char *str_display = "";
static char *str_geometry = "";
static char *str_background = "Wheat";
static char *str_foreground = "Black";
static Window scrlwin[4];
static int imgwin_noexpose = 1;
static int scrl_dir = 0;
COL color_buf[16*16];
static char grid_dash_list[2] = {1,1};

usage()
{
  fprintf(stderr, "* %s %s *\n", myname, rcsrev);
  fprintf(stderr, "written by yav (HBA00622@niftyserve.or.jp UHD98984@pcvan.or.jp)\n");
  fprintf(stderr, "usage : %s [options] col.kcf rei.cel\n", myname);
  exit(1);
}

char *filename(name)
     char *name;
{
  char *p;
  extern char *strrchr();

  if (name[0] == '\0')
    return "*Noname*";
  p = strrchr(name, '/');
  if (p != NULL)
    return p+1;
  return name;
}

#define col8to32(x) (((x)<<24)|((x)<<16)|((x)<<8)|(x))
get_color(p, n)
     COL *p;
{
  XColor col;
  int i, r;

  for (i = 0; i < n; i++) {
    col.red = col8to32(p->red);
    col.green = col8to32(p->green);
    col.blue = col8to32(p->blue);
    if (XAllocColor(dsp, DefaultColormap(dsp, scr), &col)) {
      alloced_color++;
    } else {
      fprintf(stderr, "color %d %d %d cannot allocated!\n",
	      p->red, p->green, p->blue);
      col.pixel = bg;
    }
    pixtbl[i] = col.pixel;
    XSetForeground(dsp, colgc[i], col.pixel);
    XClearArea(dsp, colwin[i], 0, 0, 0, 0, True);
    p++;
  }
  return i;
}

free_color()
{
  if (alloced_color) {
    XFreeColors(dsp, DefaultColormap(dsp, scr), pixtbl, alloced_color, 0);
    alloced_color = 0;
  }
}

init_image()
{
  if (!alloced_color)
    get_color(color_buf+16*current_pal, MAXCOLSET);
  memset(img->data, 0, img->bytes_per_line * img->height);
}

void draw_grid(x0, y0, w, h)
{
  int n, x, y;

  n = grid_size*zoomfactor;
  if (h > 0) {
    for (y = y0; y < y0 + h; y++) {
      if (((imgofsy + y)%n) == 0)
	XDrawLine(dsp, imgwin, gridgc, x0, y, x0+w-1, y);
    }
  }
  if (w > 0) {
    for (x = x0; x < x0 + w; x++) {
      if (((imgofsx + x)%n) == 0)
	XDrawLine(dsp, imgwin, gridgc, x, y0, x, y0+h-1);
    }
  }
}

void img_expose(x, y, w, h)
{
  int i;
  
  if ((imgofsx+x >= imgfilew*zoomfactor)||
      (imgofsy+y >= imgfileh*zoomfactor)||
      (imgofsx+x+w < 0)||
      (imgofsy+y+h < 0)) {
    XClearArea(dsp, imgwin, x, y, w, h, False);
    return;
  }
  i = -imgofsx - x;
  if (i > 0) {
    XClearArea(dsp, imgwin, x, y, i, h, False);
    x += i;
    w -= i;
  }
  i = -imgofsy - y;
  if (i > 0) {
    XClearArea(dsp, imgwin, x, y, w, i, False);
    y += i;
    h -= i;
  }
  i = x + w - (imgfilew*zoomfactor - imgofsx);
  if (i > 0) {
    XClearArea(dsp, imgwin, x+w-i, y, i, h, False);
    w -= i;
  }
  i = y + h - (imgfileh*zoomfactor - imgofsy);
  if (i > 0) {
    XClearArea(dsp, imgwin, x, y+h-i, w, i, False);
    h -= i;
  }
  if (w&&h) {
    XPutImage(dsp, imgwin, gc0, img, x+imgofsx, y+imgofsy, x, y, w, h);
    if (grid_mode)
      draw_grid(x, y, w, h);
  }
}

void image_update_all()
{
  gen_image_sub(0, 0, imgfilew, imgfileh);
  img_expose(0, 0, imgwinwidth, imgwinheight);
}

void color_update_all()
{
  int i;
  
  free_color();
  get_color(color_buf+16*current_pal, MAXCOLSET);
  image_update_all();
}

init_color_buf()
{
  int i;
  static COL initdata[16] = {
    {  0,  0,  0},{  0,  0,255},{255,  0,  0},{255,  0,255},
    {  0,255,  0},{  0,255,255},{255,255,  0},{127,127,127},
    { 64, 64, 64},{127,127,255},{255,127,127},{255,127,255},
    {127,255,127},{127,255,255},{255,255,127},{255,255,255}
  };
  
  for (i = 0; i < 16; i++)
    memcpy(color_buf+i*16, initdata, sizeof(COL)*16);
}

image_alloc()
{
  imgwidth = MAXIMGW;
  imgheight = MAXIMGH;
  imgdata = (unsigned char *)malloc(imgwidth*imgheight);
  if (imgdata == NULL) {
    fprintf(stderr, "more core!\n");
    exit(1);
  }
  img = XCreateImage(dsp, DefaultVisual(dsp, scr), DefaultDepth(dsp, scr),
		     ZPixmap, 0, None,
		     imgwidth*MAXZOOM, imgheight*MAXZOOM,
		     32, 0);
  if (img == NULL) {
    fprintf(stderr, "more core!");
    exit(1);
  }
  img->data = (char *)malloc(img->bytes_per_line * img->height);
  if (img->data == NULL) {
    fprintf(stderr, "more core!\n");
    exit(1);
  }
}

resize_scrl()
{
  XMoveResizeWindow(dsp, scrlwin[0], IMGX, IMGY-IMGSCRLW-1, imgwinwidth, IMGSCRLW);
  XMoveResizeWindow(dsp, scrlwin[1], IMGX, IMGY+imgwinheight+1, imgwinwidth, IMGSCRLW);
  XMoveResizeWindow(dsp, scrlwin[2], IMGX-IMGSCRLW-1, IMGY, IMGSCRLW, imgwinheight);
  XMoveResizeWindow(dsp, scrlwin[3], IMGX+imgwinwidth+1, IMGY, IMGSCRLW, imgwinheight);
}

resize_img_window()
{
  XWindowAttributes attr;

  XGetWindowAttributes(dsp, win, &attr);
  imgwinwidth = attr.width - IMGX - IMGSCRLW - 16;
  if (imgwinwidth > IMGWINWIDTHMAX)
    imgwinwidth = IMGWINWIDTHMAX;
  imgwinheight = attr.height - IMGY - IMGSCRLW - 16;
  if (imgwinheight > IMGWINHEIGHTMAX)
    imgwinheight = IMGWINHEIGHTMAX;
  XResizeWindow(dsp, imgwin, imgwinwidth, imgwinheight);
  resize_scrl();
}

create_img_window()
{
  int i;

  imgwinwidth = 32;
  imgwinheight = 32;
  for (i = 0; i < 4; i++)
    scrlwin[i] = XCreateSimpleWindow(dsp, win, IMGX, IMGY, 1, 1, 1, fg, bg);
  imgwin = XCreateSimpleWindow(dsp, win, IMGX, IMGY,
			       imgwinwidth, imgwinheight, 1, fg, bg);
  resize_scrl();
  XSelectInput(dsp, imgwin, ExposureMask|OwnerGrabButtonMask|
	       ButtonPressMask|ButtonReleaseMask|ButtonMotionMask);
  XMapRaised(dsp, imgwin);
  for (i = 0; i < 4; i++) {
    XSelectInput(dsp, scrlwin[i], ExposureMask|
		 EnterWindowMask|LeaveWindowMask|
		 ButtonPressMask|ButtonReleaseMask);
    XMapRaised(dsp, scrlwin[i]);
  }
}

close_img_window()
{
  int i;
  
  XDestroyWindow(dsp, imgwin);
  for (i = 0; i < 4; i++)
    XDestroyWindow(dsp, scrlwin[i]);
}

static int imgscrl_dir_mask[4] = {1, 2, 4, 8};

imgscrl_expose(ev)
     XEvent *ev;
{
  int i;
  static char *name[4] = {"^", "v", "<", ">"};
  
  for (i = 0; i < 4; i++) {
    if (ev->xexpose.window == scrlwin[i]) {
      menuwin_center(scrlwin[i], name[i], scrl_dir & imgscrl_dir_mask[i]);
      break;
    }
  }
}

img_scroll(dx, dy)
{
  if (dx) {
    imgofsx += dx;
    XClearArea(dsp, cursorwin[0], 0, 0, 0, 0, True);
  }
  if (dy) {
    imgofsy += dy;
    XClearArea(dsp, cursorwin[1], 0, 0, 0, 0, True);
  }
  if (imgwin_noexpose) {
    XCopyArea(dsp, imgwin, imgwin, gc0,
	      (dx<0?0:dx), (dy<0?0:dy),
	      imgwinwidth-(dx<0?-dx:dx), imgwinheight-(dy<0?-dy:dy),
	      (dx<0?-dx:0), (dy<0?-dy:0));
    imgwin_noexpose = 0;
    if (dx)
      img_expose((dx<0?0:imgwinwidth-dx), 0, (dx<0?-dx:dx), imgwinheight);
    if (dy)
      img_expose(0, (dy<0?0:imgwinheight-dy), imgwinwidth, (dy<0?-dy:dy));
  } else {
    img_expose(0, 0, imgwinwidth, imgwinheight);
  }
}

img_button(ox, oy, st)
{
  int x, y, dx, dy;
  unsigned long pix;
  static int bstat = 0;

  switch (st) {
  case 1:
    bstat |= 1;
    break;
  case 0x81:
    bstat &= ~1;
    break;
  case 2:
    bstat |= 2;
    break;
  case 0x82:
    bstat &= ~2;
    break;
  case 3:
    bstat |= 4;
    break;
  case 0x83:
    bstat &= ~4;
    break;
  }
  if (ox < 0 || ox >= imgwinwidth ||
      oy < 0 || oy >= imgwinheight)
    return;
  x = ox + imgofsx;
  y = oy + imgofsy;
  x /= zoomfactor;
  y /= zoomfactor;
  if (x != curx) {
    curx = x;
    XClearArea(dsp, cursorwin[2], 0, 0, 0, 0, True);
  }
  if (y != cury) {
    cury = y;
    XClearArea(dsp, cursorwin[3], 0, 0, 0, 0, True);
  }
  if (x < 0 || x >= imgfilew || y < 0 || y >= imgfileh)
    return;
  if (bstat & 1) {
    if (current_col >= 0) {
      setpixelcolor(x, y, current_col);
      img_expose(ox-(ox%zoomfactor), oy-(oy%zoomfactor), zoomfactor*2, zoomfactor*2);
    }
  }
  if (bstat & (2|4)) {
    int oldcol;
    
    oldcol = current_col;
    current_col = *(imgdata+y*imgwidth+x);
    if (current_col != oldcol) {
      if (oldcol >= 0)
	XClearArea(dsp, colwin[oldcol], 0, 0, 0, 0, True);
      if (current_col >= 0)
	XClearArea(dsp, colwin[current_col], 0, 0, 0, 0, True);
    }
  }
}

imgscrl_press(ev)
     XEvent *ev;
{
  int i;
  
  for (i = 0; i < 4; i++) {
    if (ev->xbutton.window == scrlwin[i]) {
      scrl_dir |= imgscrl_dir_mask[i];
      XClearArea(dsp, scrlwin[i], 0, 0, 0, 0, True);
      break;
    }
  }
}

imgscrl_release(ev)
     XEvent *ev;
{
  int i;
  
  for (i = 0; i < 4; i++) {
    if (ev->xbutton.window == scrlwin[i]) {
      scrl_dir &= ~imgscrl_dir_mask[i];
      XClearArea(dsp, scrlwin[i], 0, 0, 0, 0, True);
      break;
    }
  }
}

imgscrl_enter(ev)
     XEvent *ev;
{
  int i;
  
  for (i = 0; i < 4; i++) {
    if (ev->xcrossing.window == scrlwin[i]) {
      menuwin_highlight(scrlwin[i], scrl_dir & imgscrl_dir_mask[i]);
      break;
    }
  }
}

imgscrl_background()
{
  if (scrl_dir & imgscrl_dir_mask[0])
    img_scroll(0, -8);
  if (scrl_dir & imgscrl_dir_mask[1])
    img_scroll(0, 8);
  if (scrl_dir & imgscrl_dir_mask[2])
    img_scroll(-8, 0);
  if (scrl_dir & imgscrl_dir_mask[3])
    img_scroll(8, 0);
}

key_event(ev)
     XEvent *ev;
{
  int i;
  char buf[32];
  KeySym key;
  XComposeStatus cs;
  
  i = XLookupString((XKeyEvent *)ev, buf, sizeof(buf)-1, &key, &cs);
  buf[i] = '\0';
  switch(buf[0]) {
  case 0x03:
  case 'q':
    menu_stat_change(0);
    break;
  case 'f':
    menu_stat_change(1);
    break;
  case 'o':
    zoom_out();
    break;
  case 'i':
    zoom_in();
    break;
  case 'L'-0x40:
    XClearArea(dsp, imgwin, 0, 0, 0, 0, True);
    break;
  case '2':
  case 'j':
    img_scroll(0, 8);
    break;
  case '8':
  case 'k':
    img_scroll(0, -8);
    break;
  case '6':
  case 'l':
    img_scroll(8, 0);
    break;
  case '4':
  case 'h':
    img_scroll(-8, 0);
    break;
  case '1':
    img_scroll(-8, 8);
    break;
  case '3':
    img_scroll(8, 8);
    break;
  case '7':
    img_scroll(-8, -8);
    break;
  case '9':
    img_scroll(8, -8);
    break;
  }
}

f_display(ac, av)
     int *ac; char ***av;
{
  if (*ac > 1) {
    --*ac;
    ++*av;
    str_display = **av;
  }
}

f_geometry(ac, av)
     int *ac; char ***av;
{
  if (*ac > 1) {
    --*ac;
    ++*av;
    str_geometry = **av;
  }
}

f_background(ac, av)
     int *ac; char ***av;
{
  if (*ac > 1) {
    --*ac;
    ++*av;
    str_background = **av;
  }
}

f_foreground(ac, av)
     int *ac; char ***av;
{
  if (*ac > 1) {
    --*ac;
    ++*av;
    str_foreground = **av;
  }
}

struct {
  char *name;
  int (*func)();
} opttbl[] = {
  {"-display", f_display},
  {"-geometry", f_geometry},
  {"-background", f_background},
  {"-foreground", f_foreground},
  {"-bg", f_background},
  {"-fg", f_foreground},
  {"-?", usage},
  {"-h", usage},
  {"-help", usage},
  {NULL, NULL}
};

main(argc, argv)
     int argc; char **argv;
{
  XSizeHints hint;
  XEvent ev;
  char buf[256];
  int oargc;
  char **oargv;
  int i;
  Window double_click_win = 0;
  Time l, double_click_time;
  XColor col;

  oargc = argc;
  oargv = argv;
  myname = filename(*argv);
  while (--argc && **++argv == '-') {
    for (i = 0; opttbl[i].name != NULL; i++) {
      if (strcmp(*argv, opttbl[i].name) == 0) {
	opttbl[i].func(&argc, &argv);
	break;
      }
    }
  }
  dsp = XOpenDisplay(str_display);
  if (!dsp) {
    fprintf(stderr, "Cannot open display!\n");
    exit(1);
  }
  scr = DefaultScreen(dsp);
  if (!XParseColor(dsp, DefaultColormap(dsp, scr), str_background, &col)||
      !XAllocColor(dsp, DefaultColormap(dsp, scr), &col)) {
    fprintf(stderr, "``%s'' color alloc error!\n", str_background);
    exit(1);
  }
  bg = col.pixel;
  if (!XParseColor(dsp, DefaultColormap(dsp, scr), str_foreground, &col)||
      !XAllocColor(dsp, DefaultColormap(dsp, scr), &col)) {
    fprintf(stderr, "``%s'' color alloc error!\n", str_foreground);
    exit(1);
  }
  fg = col.pixel;
  hint.flags = 0;
  i = XGeometry(dsp, scr, str_geometry, "440x380", 2, 1, 1, 0, 0,
		&(hint.x), &(hint.y), &(hint.width), &(hint.height));
  if (i & (XValue|YValue))
    hint.flags |= PPosition;
  if (i & (WidthValue|HeightValue))
    hint.flags |= PSize;
  win = XCreateSimpleWindow(dsp, DefaultRootWindow(dsp),
			    hint.x, hint.y, hint.width, hint.height,
			    2, fg, bg);
  XSetStandardProperties(dsp, win, *oargv, *oargv, None, oargv, oargc, &hint);

  fs0 = XLoadQueryFont(dsp, "8x13");
  gc0 = XCreateGC(dsp, win, 0, 0);
  XSetBackground(dsp, gc0, bg);
  XSetForeground(dsp, gc0, fg);
  XSetFont(dsp, gc0, fs0->fid);

  fs1 = XLoadQueryFont(dsp, "8x13bold");
  gc1 = XCreateGC(dsp, win, 0, 0);
  XSetBackground(dsp, gc1, fg);
  XSetForeground(dsp, gc1, bg);
  XSetFont(dsp, gc1, fs1->fid);

  gridgc = XCreateGC(dsp, win, 0, 0);
  XSetBackground(dsp, gridgc, bg);
  XSetForeground(dsp, gridgc, fg);
  XSetLineAttributes(dsp, gridgc, 0, LineOnOffDash, CapButt, JoinMiter);
  XSetDashes(dsp, gridgc, 0, grid_dash_list, 2);

  create_img_window();

  create_menu_window(MENUX, MENUY);
  create_colset_window(COLSETX, COLSETY);
  create_palset_window(PALSETX, PALSETY);
  create_filename_window();
  create_zoom_window(256, 4);
  create_grid_window(340, 4);
  create_cursor_window(256, 32);
  image_alloc();
  
  init_color_buf();
  init_image();
  image_update_all();
  while (argc--)
    read_data_file(*argv++);
  XSelectInput(dsp, win, ExposureMask|StructureNotifyMask|
	       ButtonPressMask|KeyPressMask);
  XMapRaised(dsp, win);
  while (!sw_quit) {
    if (XEventsQueued(dsp, QueuedAfterFlush) == 0) {
      imgscrl_background();
      file_background();
    }
    XNextEvent(dsp, &ev);
    switch(ev.type) {
    case NoExpose:
      if (ev.xexpose.window == imgwin)
	imgwin_noexpose = 1;
      break;
    case GraphicsExpose:
      if (ev.xgraphicsexpose.drawable == imgwin) {
	img_expose(ev.xgraphicsexpose.x, ev.xgraphicsexpose.y,
		   ev.xgraphicsexpose.width, ev.xgraphicsexpose.height);
	if (ev.xgraphicsexpose.count == 0)
	  imgwin_noexpose = 1;
      }
      break;
    case Expose:
      if (ev.xexpose.window == win) {
      } else if (ev.xexpose.window == imgwin) {
	img_expose(ev.xexpose.x, ev.xexpose.y,
		   ev.xexpose.width, ev.xexpose.height);
      } else {
	filename_expose(&ev);
	imgscrl_expose(&ev);
	menu_expose(&ev);
	colset_expose(&ev);
	palset_expose(&ev);
	file_expose(&ev);
	zoom_expose(&ev);
	grid_expose(&ev);
	cursor_expose(&ev);
	info_expose(&ev);
      }
      break;
    case MappingNotify:
      XRefreshKeyboardMapping((XMappingEvent *)&ev);
      break;
    case KeyPress:
      key_event(&ev);
      break;
    case ConfigureNotify:
      if (ev.xconfigure.window == win)
	resize_img_window();
      break;
    case MotionNotify:
      if (ev.xmotion.window == imgwin) {
	img_button(ev.xmotion.x, ev.xmotion.y, 0);
      }
      break;
    case ButtonPress:
      if (ev.xbutton.window == double_click_win) {
	l = ev.xbutton.time - double_click_time;
	if (l & 0x80000000)
	  l = ~l + 1;
	double_clicked = l < 500;
      } else {
	double_clicked = 0;
      }
      double_click_win = ev.xbutton.window;
      double_click_time = ev.xbutton.time;
      if (ev.xbutton.window == win) {
	switch(ev.xbutton.button) {
	case 1:
	  break;
	}
      } else if (ev.xbutton.window == imgwin) {
	img_button(ev.xbutton.x, ev.xbutton.y, ev.xbutton.button);
      }
      imgscrl_press(&ev);
      menu_press(&ev);
      colset_button(&ev);
      palset_button(&ev);
      file_button(&ev);
      zoom_press(&ev);
      grid_press(&ev);
      cursor_press(&ev);
      info_press(&ev);
      break;
    case ButtonRelease:
      if (ev.xbutton.window == imgwin) {
	img_button(ev.xbutton.x, ev.xbutton.y, 0x80|ev.xbutton.button);
      }
      imgscrl_release(&ev);
      file_release(&ev);
      break;
    case EnterNotify:
      double_click_win = 0;
      imgscrl_enter(&ev);
      menu_enter(&ev);
      colset_enter(&ev);
      palset_enter(&ev);
      file_enter(&ev);
      zoom_enter(&ev);
      grid_enter(&ev);
      cursor_enter(&ev);
      info_enter(&ev);
      break;
    case LeaveNotify:
      double_click_win = 0;
      XClearArea(dsp, ev.xcrossing.window, 0, 0, 0, 0, True);
      break;
    }
  }
  close_info_window();
  close_cursor_window();
  close_grid_window();
  close_zoom_window();
  close_filename_window();
  close_file_window();
  close_palset_window();
  close_colset_window();
  close_menu_window();
  close_img_window();
  free_color();
  XFree((char *)img);
  XFreeGC(dsp, gc1);
  XFreeGC(dsp, gc0);
  XDestroyWindow(dsp, win);
  XCloseDisplay(dsp);
  exit(0);
}

/* End of file */
